__title__ = "nucypher-ops"

__url__ = "https://github.com/nucypher/nucypher-ops"

__summary__ = 'Install and management tools for a proxy re-encryption network to empower privacy in decentralized systems.'

__version__ = "0.10.0"

__author__ = "NuCypher"

__email__ = "dev@nucypher.com"

__license__ = "GNU Affero General Public License, Version 3"

__copyright__ = 'Copyright (C) 2022 NuCypher'
